@@ -56,6 +56,7 @@ on 5 January 2021.
 use std::{
     collections::{HashMap, VecDeque},
     convert::{TryFrom, TryInto},
+    fmt::Display,
     u32,
 };
 
@@ -68,6 +69,41 @@ const MAX_VERTEX_CACHE_SIZE: u16 = std::u16::MAX - 1;
 const NOT_IN_CACHE: u16 = std::u16::MAX;
 const NULL_TRI: u32 = std::u32::MAX;
 
+/// An error or invariant violation during an ordering operation
+#[derive(Debug, PartialEq)]
+pub enum Error {
+    /// Provided slice of Index cannot represent one or more index triples.
+    IndicesNotTriples,
+    /// An Index failed to convert to usize.
+    IndexToUsizeConversion,
+    /// There are more than `limit` triangles involved.
+    TooManyTrianglesInTotal { limit: usize },
+    /// More than `limit` triangles connect to vertex `vertex_idx`.
+    TooManyTrianglesAtVertex { vertex_idx: usize, limit: usize },
+    /// The resulting ordered Index draw list was malformed.
+    MalformedDrawList,
+}
+
+impl Display for Error {
+    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
+        match self {
+            Error::IndicesNotTriples => f.write_str("elements are not all triples"),
+            Error::IndexToUsizeConversion => f.write_str("can not convert Index to usize"),
+            Error::TooManyTrianglesInTotal { limit } => f.write_fmt(format_args!(
+                "too many triangles in total. {} triangles are supported",
+                limit,
+            )),
+            Error::TooManyTrianglesAtVertex { vertex_idx, limit } => f.write_fmt(format_args!(
+                "too many triangles connected to vertex {}. {} triangles are supported",
+                vertex_idx, limit,
+            )),
+            Error::MalformedDrawList => {
+                f.write_str("the generated ordered Index draw list is malformed")
+            }
+        }
+    }
+}
+
 /*
 TF: Instead, each vertex holds the following data:
 
@@ -156,9 +192,9 @@ fn calculate_vertex_score(num_active_tris: u32, cache_pos: u16, vertex_cache_siz
 /// use forsyth::order_triangles_inplace;
 ///
 /// let mut indices = [0_u16, 1, 2, 3, 2, 0, 0, 1, 3, 0, 1, 4, 0, 1, 5];
-/// assert_eq!(
-///     order_triangles_inplace(&mut indices, 32).unwrap(),
-///     &[0, 1, 4, 0, 1, 5, 0, 1, 2, 3, 2, 0, 0, 1, 3]
+/// assert!(
+///     order_triangles_inplace(&mut indices, 32) ==
+///     Ok(&[0_u16, 1, 4, 0, 1, 5, 0, 1, 2, 3, 2, 0, 0, 1, 3])
 /// );
 /// ```
 ///
@@ -167,9 +203,9 @@ fn calculate_vertex_score(num_active_tris: u32, cache_pos: u16, vertex_cache_siz
 pub fn order_triangles_inplace<Index>(
     indices: &mut [Index],
     vertex_cache_size: u16,
-) -> Result<&[Index], &str>
+) -> Result<&[Index], Error>
 where
-    Index: std::convert::TryInto<usize>
+    Index: TryInto<usize>
         + TryFrom<usize>
         + std::clone::Clone
         + std::default::Default
@@ -180,7 +216,7 @@ where
         std::cmp::max(4, std::cmp::min(vertex_cache_size, MAX_VERTEX_CACHE_SIZE));
 
     if indices.is_empty() || (indices.len() % 3) != 0 {
-        return Err("elements are not all triples");
+        return Err(Error::IndexToUsizeConversion);
     }
 
     let num_tris = indices.len() / 3;
@@ -211,17 +247,25 @@ where
                 let vertex_idx: usize = if let Ok(vertex_idx) = vertex_idx.try_into() {
                     vertex_idx
                 } else {
-                    return Err("failed to convert Index to usize");
+                    return Err(Error::IndexToUsizeConversion);
                 };
 
                 if verts.len() < vertex_idx + 1 {
                     verts.resize(vertex_idx + 1, VertexInfo::default());
                 }
                 let vert: &mut VertexInfo = &mut verts[vertex_idx];
+                if vert.num_tris_active == std::u16::MAX {
+                    return Err(Error::TooManyTrianglesAtVertex {
+                        vertex_idx,
+                        limit: std::u16::MAX as usize,
+                    });
+                }
                 vert.num_tris_active += 1;
 
                 if total_vertex_tris == std::u32::MAX {
-                    return Err("too many triangles");
+                    return Err(Error::TooManyTrianglesInTotal {
+                        limit: std::usize::MAX,
+                    });
                 }
 
                 total_vertex_tris += 1;
@@ -245,17 +289,15 @@ where
         for tri_idx in 0..num_tris {
             for v in 0..3 {
                 let element = (tri_idx * 3) + v;
-                if let Ok(vertex_idx) = (*indices.get(element).unwrap()).try_into() {
-                    let vert: &mut VertexInfo = &mut verts[vertex_idx];
-                    let tri_list = &mut vertex_tri_list[vert.tri_list_ofs as usize..];
-                    tri_list[vert.num_tris_active as usize] = tri_idx as u32;
-                    if vert.num_tris_active == std::u16::MAX {
-                        return Err("too many triangles at single vertex");
-                    }
-                    vert.num_tris_active += 1;
-                } else {
-                    return Err("failed to convert Index");
-                }
+
+                let vertex_idx = indices[element]
+                    .try_into()
+                    .map_err(|_| Error::IndexToUsizeConversion)?;
+                let vert: &mut VertexInfo = &mut verts[vertex_idx];
+                let tri_list = &mut vertex_tri_list[vert.tri_list_ofs as usize..];
+                tri_list[vert.num_tris_active as usize] = tri_idx as u32;
+                // We already checked that this is below u16::MAX
+                vert.num_tris_active += 1;
             }
         }
 
@@ -301,11 +343,9 @@ where
             and the list of triangle indices in the vertex is updated appropriately.
             */
             let vertex_idx = *v;
-            let vertex_idx_usize = if let Ok(vertex_idx_usize) = Index::try_into(vertex_idx) {
-                vertex_idx_usize
-            } else {
-                return Err("failed to convert Index to usize");
-            };
+            let vertex_idx_usize: usize = vertex_idx
+                .try_into()
+                .map_err(|_| Error::IndexToUsizeConversion)?;
 
             let vert = &mut verts[vertex_idx_usize];
 
@@ -328,7 +368,7 @@ where
                 *draw_list_idx = vertex_idx;
                 draw_list_cursor += 1;
             } else {
-                return Err("draw list index out of bounds");
+                return Err(Error::MalformedDrawList);
             }
 
             /*
@@ -350,11 +390,9 @@ where
         for (cache_pos, vertex_idx) in lru_cache.iter().enumerate() {
             let cache_pos = std::cmp::min(cache_pos, (MAX_VERTEX_CACHE_SIZE - 1) as usize) as u16;
             let (new_score, old_score, tri_list_ofs, num_tris_active) = {
-                let vertex_idx = if let Ok(vertex_idx) = (*vertex_idx).try_into() {
-                    vertex_idx
-                } else {
-                    return Err("failed to convert Index into usize");
-                };
+                let vertex_idx: usize = (*vertex_idx)
+                    .try_into()
+                    .map_err(|_| Error::IndexToUsizeConversion)?;
                 let vert: &mut VertexInfo = &mut verts[vertex_idx];
                 vert.cache_pos = if cache_pos < vertex_cache_size {
                     cache_pos
@@ -432,7 +470,7 @@ where
     }
 
     if draw_list_cursor != num_tris * 3 {
-        return Err("draw list is incomplete");
+        return Err(Error::MalformedDrawList);
     }
 
     Ok(&indices[0..draw_list_cursor])
@@ -444,61 +482,61 @@ where
 /// allocated ordered index buffer.
 ///
 /// ```
-/// use forsyth::order_triangles;
+/// use forsyth::{Error,order_triangles};
 ///
 /// // a triangle
-/// assert_eq!(order_triangles(&[0, 1, 2], 32).unwrap(), [0, 1, 2]);
+/// assert_eq!(order_triangles(&[0, 1, 2], 32), Ok(vec![0, 1, 2]));
 /// // a quad
 /// assert_eq!(
-///     order_triangles(&[0_u32, 1, 2, 0, 2, 3], 32).unwrap(),
-///     [0, 1, 2, 0, 2, 3]
+///     order_triangles(&[0_u32, 1, 2, 0, 2, 3], 32),
+///     Ok(vec![0, 1, 2, 0, 2, 3]),
 /// );
 ///
 /// assert_eq!(
 ///     order_triangles(&Vec::<u32>::new(), 32),
-///     Err("elements are not all triples".to_string())
+///     Err(Error::IndicesNotTriples)
 /// );
 /// assert_eq!(
 ///     order_triangles(&[0, 1], 32),
-///     Err("elements are not all triples".to_string())
+///     Err(Error::IndicesNotTriples)
 /// );
 /// assert_eq!(
 ///     order_triangles(&[0, 1, 2, 3], 32),
-///     Err("elements are not all triples".to_string())
+///     Err(Error::IndicesNotTriples)
 /// );
 ///
 /// // Works with all (un-)signed integers
 /// assert_eq!(
-///     order_triangles(&[0_u8, 1, 2, 3, 4, 5], 32).unwrap(),
-///     [0, 1, 2, 3, 4, 5]
+///     order_triangles(&[0_u8, 1, 2, 3, 4, 5], 32),
+///     Ok(vec![0, 1, 2, 3, 4, 5])
 /// );
 /// assert_eq!(
-///     order_triangles(&[0_i16, 1, 2, 3, 4, 5], 32).unwrap(),
-///     [0, 1, 2, 3, 4, 5]
+///     order_triangles(&[0_i16, 1, 2, 3, 4, 5], 32),
+///     Ok(vec![0, 1, 2, 3, 4, 5])
 /// );
 /// assert_eq!(
-///     order_triangles(&[0_u32, 1, 2, 3, 4, 5], 32).unwrap(),
-///     [0, 1, 2, 3, 4, 5]
+///     order_triangles(&[0_u32, 1, 2, 3, 4, 5], 32),
+///     Ok(vec![0, 1, 2, 3, 4, 5])
 /// );
 /// assert_eq!(
-///     order_triangles(&[0_i64, 1, 2, 3, 4, 5], 32).unwrap(),
-///     [0, 1, 2, 3, 4, 5]
+///     order_triangles(&[0_i64, 1, 2, 3, 4, 5], 32),
+///     Ok(vec![0, 1, 2, 3, 4, 5])
 /// );
 /// assert_eq!(
-///     order_triangles(&[0_usize, 1, 2, 3, 4, 5], 32).unwrap(),
-///     [0, 1, 2, 3, 4, 5]
+///     order_triangles(&[0_usize, 1, 2, 3, 4, 5], 32),
+///     Ok(vec![0, 1, 2, 3, 4, 5])
 /// );
 ///
 /// // Indices can be sparse
 /// assert_eq!(
-///     order_triangles(&[0_i8, 1, 2, 63, 64, 65], 32).unwrap(),
-///     [0, 1, 2, 63, 64, 65]
+///     order_triangles(&[0_i8, 1, 2, 63, 64, 65], 32),
+///     Ok(vec![0, 1, 2, 63, 64, 65])
 /// );
 ///
 /// // Indices cannot be negative
 /// assert_eq!(
 ///    order_triangles(&[0_i8, -1, -2, 63, 64, 65], 32),
-///    Err("error ordering triangles: failed to convert Index to usize".to_string())
+///    Err(Error::IndexToUsizeConversion)
 /// );
 /// ```
 ///
@@ -507,7 +545,7 @@ where
 pub fn order_triangles<Index>(
     indices: &[Index],
     vertex_cache_size: u16,
-) -> Result<Vec<Index>, String>
+) -> Result<Vec<Index>, Error>
 where
     Index: std::convert::TryInto<usize>
         + std::convert::TryFrom<usize>
@@ -517,18 +555,15 @@ where
         + Copy,
 {
     if indices.is_empty() || (indices.len() % 3) != 0 {
-        return Err("elements are not all triples".to_string());
+        return Err(Error::IndicesNotTriples);
     }
 
     let buffer = {
         let mut buffer = Vec::with_capacity(indices.len());
         buffer.extend_from_slice(indices);
 
-        let res = order_triangles_inplace(buffer.as_mut_slice(), vertex_cache_size);
+        order_triangles_inplace(buffer.as_mut_slice(), vertex_cache_size)?;
 
-        if let Err(err) = res {
-            return Err(format!("error ordering triangles: {}", err));
-        }
         buffer
     };
     Ok(buffer)
@@ -583,18 +618,15 @@ where
         let mapped_idx = match index_map.entry(*index) {
             std::collections::hash_map::Entry::Occupied(mapped) => *mapped.get(),
             std::collections::hash_map::Entry::Vacant(vacant) => {
-                let index_usize: usize = if let Ok(index_usize) = (*index).try_into() {
-                    index_usize
-                } else {
-                    return Err("failed to convert Index to usize".to_string());
-                };
-                if let Some(vertex) = vertices.get(index_usize) {
+                let index = (*index)
+                    .try_into()
+                    .map_err(|_| "failed to convert Index to usize".to_string())?;
+                if let Some(vertex) = vertices.get(index) {
                     ordered_vertices.push(*vertex);
-                    if let Ok(mapped) = (ordered_vertices.len() - 1).try_into() {
-                        *vacant.insert(mapped)
-                    } else {
-                        return Err("failed to convert usize to Index".to_string());
-                    }
+                    let mapped = (ordered_vertices.len() - 1)
+                        .try_into()
+                        .map_err(|_| "error mapping index".to_string())?;
+                    *vacant.insert(mapped)
                 } else {
                     return Err(
                         "index does not point into vertex array (out-of-bounds)".to_string()
@@ -634,7 +666,7 @@ mod tests {
         {
             assert_eq!(
                 order_triangles(&[i16::from_be_bytes([130_u8, 246])], 9),
-                Err("elements are not all triples".to_string())
+                Err(Error::IndicesNotTriples)
             );
         }
     }
@@ -717,6 +749,16 @@ mod tests {
             );
         }
 
+        assert_eq!(
+            order_triangles_inplace::<u32>(&mut [], 32),
+            Err(Error::IndexToUsizeConversion)
+        );
+
+        assert_eq!(
+            order_triangles_inplace(&mut [-1_i32, 0, 2], 32),
+            Err(Error::IndexToUsizeConversion)
+        );
+
         // 1 shared edge - quad
         assert_eq!(
             order_triangles(&[0_u32, 1, 2, 0, 2, 3], 32).unwrap(),
@@ -725,20 +767,14 @@ mod tests {
 
         assert_eq!(
             order_triangles(&Vec::<u32>::new(), 32),
-            Err("elements are not all triples".to_string())
-        );
-        assert_eq!(
-            order_triangles(&[0], 32),
-            Err("elements are not all triples".to_string())
-        );
-        assert_eq!(
-            order_triangles(&[0, 1], 32),
-            Err("elements are not all triples".to_string())
+            Err(Error::IndicesNotTriples)
         );
+        assert_eq!(order_triangles(&[0], 32), Err(Error::IndicesNotTriples));
+        assert_eq!(order_triangles(&[0, 1], 32), Err(Error::IndicesNotTriples));
         assert_eq!(order_triangles(&[0, 1, 2], 32).unwrap(), [0, 1, 2]);
         assert_eq!(
             order_triangles(&[0, 1, 2, 3], 32),
-            Err("elements are not all triples".to_string())
+            Err(Error::IndicesNotTriples)
         );
 
         // Two disjunctive tris in multiple types
@@ -789,8 +825,53 @@ mod tests {
         // Indices are negative
         assert_eq!(
             order_triangles(&[0_i8, -1, -2, 63, 64, 65], 32),
-            Err("error ordering triangles: failed to convert Index to usize".to_string())
+            Err(Error::IndexToUsizeConversion)
         );
+
+        {
+            let num_indices = 3 * 256;
+            let mut indices = Vec::with_capacity(num_indices);
+            for _ in 0..num_indices {
+                indices.push(indices.len());
+            }
+
+            for cs in [
+                0,
+                1,
+                2,
+                3,
+                4,
+                8,
+                16,
+                32,
+                64,
+                128,
+                256,
+                257,
+                std::u16::MAX - 1,
+                std::u16::MAX,
+            ] {
+                let result = order_triangles(indices.as_slice(), cs);
+                assert!(result.is_ok(), "result is {:?}", result);
+            }
+        }
+
+        {
+            let mut indices = Vec::with_capacity((std::u16::MAX as usize) + 1);
+            for _ in 0..(std::u16::MAX as usize) + 1 {
+                indices.push(0);
+                indices.push(1);
+                indices.push(2);
+            }
+
+            assert_eq!(
+                order_triangles(indices.as_slice(), 32),
+                Err(Error::TooManyTrianglesAtVertex {
+                    vertex_idx: 0,
+                    limit: 65535
+                })
+            );
+        }
     }
 
     #[test]
@@ -804,5 +885,15 @@ mod tests {
             order_vertices(&['x', 'x', 'a', 'b', 'y', 'c'], &[2, 3, 5]),
             Ok((vec!['a', 'b', 'c'], vec![0, 1, 2]))
         );
+
+        assert_eq!(
+            order_vertices(&['a', 'b', 'c'], &[-1, 0, 2]),
+            Err("failed to convert Index to usize".to_string())
+        );
+
+        assert_eq!(
+            order_vertices(&['a', 'b', 'c'], &[0, 1, 3]),
+            Err("index does not point into vertex array (out-of-bounds)".to_string())
+        );
     }
 }
